@@ -1,4 +1,4 @@
-use std::convert::TryFrom;
+use std::fmt;
 
 use afarray::{Array, ArrayExt, Coords, Offsets};
 use async_trait::async_trait;
@@ -6,6 +6,7 @@ use futures::future::{self, TryFutureExt};
 use futures::stream::{self, StreamExt, TryStreamExt};
 use futures::try_join;
 use log::debug;
+use safecast::AsType;
 
 use tc_btree::*;
 use tc_error::*;
@@ -16,16 +17,20 @@ use tcgeneric::{TCBoxStream, TCBoxTryFuture, TCBoxTryStream};
 
 use crate::sparse::{SparseAccess, SparseAccessor};
 use crate::stream::{Read, ReadValueAt};
-use crate::{transform, Bounds, Coord, Phantom, Shape, TensorAccess, TensorType, ERR_INF, ERR_NAN};
+use crate::{
+    transform, AxisBounds, Bounds, Coord, Phantom, Shape, TensorAccess, TensorType, ERR_INF,
+    ERR_NAN,
+};
 
 use super::file::{BlockListFile, BlockListFileSlice};
 use super::stream::SparseValueStream;
 use super::{DenseTensor, PER_BLOCK};
+use std::fmt::{Formatter, Write};
 
 /// Common [`DenseTensor`] access methods
 #[async_trait]
 pub trait DenseAccess<FD: File<Array>, FS: File<Node>, D: Dir, T: Transaction<D>>:
-    Clone + ReadValueAt<D, Txn = T> + TensorAccess + Send + Sync + Sized + 'static
+    Clone + ReadValueAt<D, Txn = T> + TensorAccess + fmt::Display + Send + Sync + Sized + 'static
 {
     /// The type returned by `slice`
     type Slice: DenseAccess<FD, FS, D, T>;
@@ -38,6 +43,8 @@ pub trait DenseAccess<FD: File<Array>, FS: File<Node>, D: Dir, T: Transaction<D>
 
     /// Return a stream of the [`Array`]s which this [`DenseTensor`] comprises.
     fn block_stream<'a>(self, txn: Self::Txn) -> TCBoxTryFuture<'a, TCBoxTryStream<'a, Array>> {
+        debug!("DenseAccess::block_stream");
+
         Box::pin(async move {
             let blocks = self.value_stream(txn).await?;
             let blocks = blocks
@@ -52,6 +59,8 @@ pub trait DenseAccess<FD: File<Array>, FS: File<Node>, D: Dir, T: Transaction<D>
 
     /// Return a stream of the elements of this [`DenseTensor`].
     fn value_stream<'a>(self, txn: Self::Txn) -> TCBoxTryFuture<'a, TCBoxTryStream<'a, Number>> {
+        debug!("DenseAccess::value_stream");
+
         Box::pin(async move {
             let values = self.block_stream(txn).await?;
 
@@ -99,6 +108,7 @@ pub enum DenseAccessor<FD, FS, D, T> {
     Broadcast(Box<BlockListBroadcast<FD, FS, D, T, Self>>),
     Cast(Box<BlockListCast<FD, FS, D, T, Self>>),
     Combine(Box<BlockListCombine<FD, FS, D, T, Self, Self>>),
+    Const(Box<BlockListConst<FD, FS, D, T, Self>>),
     Expand(Box<BlockListExpand<FD, FS, D, T, Self>>),
     File(BlockListFile<FD, FS, D, T>),
     Reduce(Box<BlockListReduce<FD, FS, D, T, Self>>),
@@ -110,10 +120,11 @@ pub enum DenseAccessor<FD, FS, D, T> {
 
 impl<FD, FS, D, T> TensorAccess for DenseAccessor<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
 {
     fn dtype(&self) -> NumberType {
@@ -121,6 +132,7 @@ where
             Self::Broadcast(broadcast) => broadcast.dtype(),
             Self::Cast(cast) => cast.dtype(),
             Self::Combine(combine) => combine.dtype(),
+            Self::Const(combine) => combine.dtype(),
             Self::Expand(expansion) => expansion.dtype(),
             Self::File(file) => file.dtype(),
             Self::Reduce(reduced) => reduced.dtype(),
@@ -136,6 +148,7 @@ where
             Self::Broadcast(broadcast) => broadcast.ndim(),
             Self::Cast(cast) => cast.ndim(),
             Self::Combine(combine) => combine.ndim(),
+            Self::Const(combine) => combine.ndim(),
             Self::Expand(expansion) => expansion.ndim(),
             Self::File(file) => file.ndim(),
             Self::Reduce(reduced) => reduced.ndim(),
@@ -151,6 +164,7 @@ where
             Self::Broadcast(broadcast) => broadcast.shape(),
             Self::Cast(cast) => cast.shape(),
             Self::Combine(combine) => combine.shape(),
+            Self::Const(combine) => combine.shape(),
             Self::Expand(expansion) => expansion.shape(),
             Self::File(file) => file.shape(),
             Self::Reduce(reduced) => reduced.shape(),
@@ -166,6 +180,7 @@ where
             Self::Broadcast(broadcast) => broadcast.size(),
             Self::Cast(cast) => cast.size(),
             Self::Combine(combine) => combine.size(),
+            Self::Const(combine) => combine.size(),
             Self::Expand(expansion) => expansion.size(),
             Self::File(file) => file.size(),
             Self::Reduce(reduced) => reduced.size(),
@@ -180,10 +195,11 @@ where
 #[async_trait]
 impl<FD, FS, D, T> DenseAccess<FD, FS, D, T> for DenseAccessor<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
 {
     type Slice = Self;
@@ -199,6 +215,7 @@ where
             Self::Slice(slice) => slice.block_stream(txn),
             Self::Broadcast(broadcast) => broadcast.block_stream(txn),
             Self::Cast(cast) => cast.block_stream(txn),
+            Self::Const(combine) => combine.block_stream(txn),
             Self::Combine(combine) => combine.block_stream(txn),
             Self::Expand(expansion) => expansion.block_stream(txn),
             Self::Reduce(reduced) => reduced.block_stream(txn),
@@ -215,6 +232,7 @@ where
             Self::Broadcast(broadcast) => broadcast.value_stream(txn),
             Self::Cast(cast) => cast.value_stream(txn),
             Self::Combine(combine) => combine.value_stream(txn),
+            Self::Const(combine) => combine.value_stream(txn),
             Self::Expand(expansion) => expansion.value_stream(txn),
             Self::Reduce(reduced) => reduced.value_stream(txn),
             Self::Sparse(sparse) => sparse.value_stream(txn),
@@ -230,6 +248,7 @@ where
             Self::Broadcast(broadcast) => broadcast.slice(bounds).map(|slice| slice.accessor()),
             Self::Cast(cast) => cast.slice(bounds).map(|slice| slice.accessor()),
             Self::Combine(combine) => combine.slice(bounds).map(|slice| slice.accessor()),
+            Self::Const(combine) => combine.slice(bounds).map(|slice| slice.accessor()),
             Self::Expand(expansion) => expansion.slice(bounds).map(|slice| slice.accessor()),
             Self::Reduce(reduced) => reduced.slice(bounds).map(|slice| slice.accessor()),
             Self::Sparse(sparse) => sparse.slice(bounds).map(|slice| slice.accessor()),
@@ -243,30 +262,42 @@ where
             Self::File(file) => file
                 .transpose(permutation)
                 .map(|transpose| transpose.accessor()),
+
             Self::Slice(slice) => slice
                 .transpose(permutation)
                 .map(|transpose| transpose.accessor()),
+
             Self::Broadcast(broadcast) => broadcast
                 .transpose(permutation)
                 .map(|transpose| transpose.accessor()),
+
             Self::Cast(cast) => cast
                 .transpose(permutation)
                 .map(|transpose| transpose.accessor()),
+
             Self::Combine(combine) => combine
                 .transpose(permutation)
                 .map(|transpose| transpose.accessor()),
+
+            Self::Const(combine) => combine
+                .transpose(permutation)
+                .map(|transpose| transpose.accessor()),
+
             Self::Expand(expansion) => expansion
                 .transpose(permutation)
                 .map(|transpose| transpose.accessor()),
+
             Self::Reduce(reduced) => reduced
                 .transpose(permutation)
                 .map(|transpose| transpose.accessor()),
+
             Self::Sparse(sparse) => sparse
                 .transpose(permutation)
                 .map(|transpose| transpose.accessor()),
             Self::Transpose(transpose) => transpose
                 .transpose(permutation)
                 .map(|transpose| transpose.accessor()),
+
             Self::Unary(unary) => unary
                 .transpose(permutation)
                 .map(|transpose| transpose.accessor()),
@@ -280,6 +311,7 @@ where
             Self::Broadcast(broadcast) => broadcast.read_values(txn, coords).await,
             Self::Cast(cast) => cast.read_values(txn, coords).await,
             Self::Combine(combine) => combine.read_values(txn, coords).await,
+            Self::Const(combine) => combine.read_values(txn, coords).await,
             Self::Expand(expansion) => expansion.read_values(txn, coords).await,
             Self::Reduce(reduced) => reduced.read_values(txn, coords).await,
             Self::Sparse(sparse) => sparse.read_values(txn, coords).await,
@@ -292,10 +324,11 @@ where
 #[async_trait]
 impl<FD, FS, D, T> DenseWrite<FD, FS, D, T> for DenseAccessor<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
 {
     async fn write<V: DenseAccess<FD, FS, D, T>>(
@@ -320,10 +353,11 @@ where
 
 impl<FD, FS, D, T> ReadValueAt<D> for DenseAccessor<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
 {
     type Txn = T;
@@ -335,6 +369,7 @@ where
             Self::Broadcast(broadcast) => broadcast.read_value_at(txn, coord),
             Self::Cast(cast) => cast.read_value_at(txn, coord),
             Self::Combine(combine) => combine.read_value_at(txn, coord),
+            Self::Const(combine) => combine.read_value_at(txn, coord),
             Self::Expand(expansion) => expansion.read_value_at(txn, coord),
             Self::Reduce(reduced) => reduced.read_value_at(txn, coord),
             Self::Sparse(sparse) => sparse.read_value_at(txn, coord),
@@ -356,6 +391,24 @@ where
     }
 }
 
+impl<FD, FS, D, T> fmt::Display for DenseAccessor<FD, FS, D, T> {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        match self {
+            Self::File(file) => fmt::Display::fmt(file, f),
+            Self::Slice(slice) => fmt::Display::fmt(slice, f),
+            Self::Broadcast(broadcast) => fmt::Display::fmt(broadcast, f),
+            Self::Cast(cast) => fmt::Display::fmt(cast, f),
+            Self::Combine(combine) => fmt::Display::fmt(combine, f),
+            Self::Const(combine) => fmt::Display::fmt(combine, f),
+            Self::Expand(expand) => fmt::Display::fmt(expand, f),
+            Self::Reduce(reduce) => fmt::Display::fmt(reduce, f),
+            Self::Sparse(sparse) => fmt::Display::fmt(sparse, f),
+            Self::Transpose(transpose) => fmt::Display::fmt(transpose, f),
+            Self::Unary(unary) => fmt::Display::fmt(unary, f),
+        }
+    }
+}
+
 #[derive(Clone)]
 pub struct BlockListCombine<FD, FS, D, T, L, R> {
     left: L,
@@ -368,10 +421,11 @@ pub struct BlockListCombine<FD, FS, D, T, L, R> {
 
 impl<FD, FS, D, T, L, R> BlockListCombine<FD, FS, D, T, L, R>
 where
-    FD: File<Array>,
-    FS: File<Node>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     L: DenseAccess<FD, FS, D, T>,
     R: DenseAccess<FD, FS, D, T>,
 {
@@ -429,13 +483,14 @@ where
 #[async_trait]
 impl<FD, FS, D, T, L, R> DenseAccess<FD, FS, D, T> for BlockListCombine<FD, FS, D, T, L, R>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
+    D::FileClass: From<TensorType>,
     L: DenseAccess<FD, FS, D, T>,
     R: DenseAccess<FD, FS, D, T>,
-    D::FileClass: From<TensorType>,
 {
     type Slice = BlockListCombine<FD, FS, D, T, L::Slice, R::Slice>;
     type Transpose = BlockListCombine<FD, FS, D, T, L::Transpose, R::Transpose>;
@@ -488,6 +543,11 @@ where
     }
 
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
+        debug!(
+            "slice {} from BlockListCombine {}, {}",
+            bounds, self.left, self.right
+        );
+
         let left = self.left.slice(bounds.clone())?;
         let right = self.right.slice(bounds)?;
 
@@ -539,13 +599,14 @@ where
 
 impl<FD, FS, D, T, L, R> ReadValueAt<D> for BlockListCombine<FD, FS, D, T, L, R>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
+    D::FileClass: From<TensorType>,
     L: DenseAccess<FD, FS, D, T>,
     R: DenseAccess<FD, FS, D, T>,
-    D::FileClass: From<TensorType>,
 {
     type Txn = T;
 
@@ -566,6 +627,159 @@ where
     }
 }
 
+impl<FD, FS, D, T, L, R> fmt::Display for BlockListCombine<FD, FS, D, T, L, R> {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        f.write_str("dense Tensor-Tensor op")
+    }
+}
+
+#[derive(Clone)]
+pub struct BlockListConst<FD, FS, D, T, B> {
+    source: B,
+    other: Number,
+    combinator: fn(Array, Number) -> Array,
+    value_combinator: fn(Number, Number) -> Number,
+    phantom: Phantom<FD, FS, D, T>,
+}
+
+impl<FD, FS, D, T, B> BlockListConst<FD, FS, D, T, B> {
+    pub fn new(
+        source: B,
+        other: Number,
+        combinator: fn(Array, Number) -> Array,
+        value_combinator: fn(Number, Number) -> Number,
+    ) -> Self {
+        debug!("BlockListConst::new");
+
+        Self {
+            source,
+            other,
+            combinator,
+            value_combinator,
+            phantom: Phantom::default(),
+        }
+    }
+}
+
+impl<FD, FS, D, T, B> TensorAccess for BlockListConst<FD, FS, D, T, B>
+where
+    D: Dir,
+    T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    B: DenseAccess<FD, FS, D, T>,
+{
+    fn dtype(&self) -> NumberType {
+        Ord::max(self.source.dtype(), self.other.class())
+    }
+
+    fn ndim(&self) -> usize {
+        self.source.ndim()
+    }
+
+    fn shape(&'_ self) -> &'_ Shape {
+        self.source.shape()
+    }
+
+    fn size(&self) -> u64 {
+        self.source.size()
+    }
+}
+
+#[async_trait]
+impl<FD, FS, D, T, B> DenseAccess<FD, FS, D, T> for BlockListConst<FD, FS, D, T, B>
+where
+    FD: File<Array>,
+    FS: File<Node>,
+    D: Dir,
+    T: Transaction<D>,
+    B: DenseAccess<FD, FS, D, T>,
+{
+    type Slice = BlockListConst<FD, FS, D, T, B::Slice>;
+    type Transpose = BlockListConst<FD, FS, D, T, B::Transpose>;
+
+    fn accessor(self) -> DenseAccessor<FD, FS, D, T> {
+        let this = BlockListConst {
+            source: self.source.accessor(),
+            other: self.other,
+            combinator: self.combinator,
+            value_combinator: self.value_combinator,
+            phantom: self.phantom,
+        };
+
+        DenseAccessor::Const(Box::new(this))
+    }
+
+    fn block_stream<'a>(self, txn: T) -> TCBoxTryFuture<'a, TCBoxTryStream<'a, Array>> {
+        Box::pin(async move {
+            let combinator = (self.combinator);
+            let right = self.other;
+
+            let left = self.source.block_stream(txn).await?;
+            let blocks = left.map_ok(move |block| combinator(block, right));
+            let blocks: TCBoxTryStream<'a, Array> = Box::pin(blocks);
+            Ok(blocks)
+        })
+    }
+
+    fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
+        let slice = self.source.slice(bounds)?;
+        Ok(BlockListConst::new(
+            slice,
+            self.other,
+            self.combinator,
+            self.value_combinator,
+        ))
+    }
+
+    fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
+        let transpose = self.source.transpose(permutation)?;
+        Ok(BlockListConst::new(
+            transpose,
+            self.other,
+            self.combinator,
+            self.value_combinator,
+        ))
+    }
+
+    async fn read_values(self, txn: Self::Txn, coords: Coords) -> TCResult<Array> {
+        let combinator = self.combinator;
+        let other = self.other;
+        self.source
+            .read_values(txn, coords)
+            .map_ok(|values| combinator(values, other))
+            .await
+    }
+}
+
+impl<FD, FS, D, T, B> ReadValueAt<D> for BlockListConst<FD, FS, D, T, B>
+where
+    D: Dir,
+    T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    B: DenseAccess<FD, FS, D, T>,
+{
+    type Txn = T;
+
+    fn read_value_at<'a>(self, txn: Self::Txn, coord: Coord) -> Read<'a> {
+        let combinator = self.value_combinator;
+        let other = self.other;
+        let read = self
+            .source
+            .read_value_at(txn, coord)
+            .map_ok(move |(coord, val)| (coord, combinator(val, other)));
+
+        Box::pin(read)
+    }
+}
+
+impl<FD, FS, D, T, B> fmt::Display for BlockListConst<FD, FS, D, T, B> {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        f.write_str("dense Tensor-constant op")
+    }
+}
+
 #[derive(Clone)]
 pub struct BlockListBroadcast<FD, FS, D, T, B> {
     source: B,
@@ -583,6 +797,7 @@ where
 {
     pub fn new(source: B, shape: Shape) -> TCResult<Self> {
         let rebase = transform::Broadcast::new(source.shape().clone(), shape)?;
+
         Ok(Self {
             source,
             rebase,
@@ -619,12 +834,13 @@ where
 #[async_trait]
 impl<FD, FS, D, T, B> DenseAccess<FD, FS, D, T> for BlockListBroadcast<FD, FS, D, T, B>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
-    B: DenseAccess<FD, FS, D, T>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
+    B: DenseAccess<FD, FS, D, T>,
 {
     type Slice = BlockListBroadcast<FD, FS, D, T, B::Slice>;
     type Transpose = BlockListTranspose<FD, FS, D, T, Self>;
@@ -665,10 +881,13 @@ where
     }
 
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
+        debug!("slice {} from broadcast of {}", bounds, self.source.shape());
         self.shape().validate_bounds(&bounds)?;
 
         let shape = bounds.to_shape(self.shape())?;
         let bounds = self.rebase.invert_bounds(bounds);
+        debug!("source bounds are {}, broadcast into {}", bounds, shape);
+
         let source = self.source.slice(bounds)?;
         BlockListBroadcast::new(source, shape)
     }
@@ -691,12 +910,13 @@ where
 
 impl<FD, FS, D, T, B> ReadValueAt<D> for BlockListBroadcast<FD, FS, D, T, B>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
-    B: DenseAccess<FD, FS, D, T>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
+    B: DenseAccess<FD, FS, D, T>,
 {
     type Txn = T;
 
@@ -711,6 +931,12 @@ where
     }
 }
 
+impl<FD, FS, D, T, B> fmt::Display for BlockListBroadcast<FD, FS, D, T, B> {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        f.write_str("dense Tensor broadcast")
+    }
+}
+
 #[derive(Clone)]
 pub struct BlockListCast<FD, FS, D, T, B> {
     source: B,
@@ -763,12 +989,13 @@ where
 #[async_trait]
 impl<FD, FS, D, T, B> DenseAccess<FD, FS, D, T> for BlockListCast<FD, FS, D, T, B>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
-    B: DenseAccess<FD, FS, D, T>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
+    B: DenseAccess<FD, FS, D, T>,
 {
     type Slice = BlockListCast<FD, FS, D, T, B::Slice>;
     type Transpose = BlockListCast<FD, FS, D, T, B::Transpose>;
@@ -810,12 +1037,13 @@ where
 
 impl<FD, FS, D, T, B> ReadValueAt<D> for BlockListCast<FD, FS, D, T, B>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
-    B: DenseAccess<FD, FS, D, T>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
+    B: DenseAccess<FD, FS, D, T>,
 {
     type Txn = T;
 
@@ -830,6 +1058,12 @@ where
     }
 }
 
+impl<FD, FS, D, T, B> fmt::Display for BlockListCast<FD, FS, D, T, B> {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        f.write_str("dense Tensor type cast")
+    }
+}
+
 #[derive(Clone)]
 pub struct BlockListExpand<FD, FS, D, T, B> {
     source: B,
@@ -883,12 +1117,13 @@ where
 #[async_trait]
 impl<FD, FS, D, T, B> DenseAccess<FD, FS, D, T> for BlockListExpand<FD, FS, D, T, B>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
-    B: DenseAccess<FD, FS, D, T>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
+    B: DenseAccess<FD, FS, D, T>,
 {
     type Slice = B::Slice;
     type Transpose = B::Transpose;
@@ -931,12 +1166,13 @@ where
 
 impl<FD, FS, D, T, B> ReadValueAt<D> for BlockListExpand<FD, FS, D, T, B>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
-    B: DenseAccess<FD, FS, D, T>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
+    B: DenseAccess<FD, FS, D, T>,
 {
     type Txn = T;
 
@@ -951,6 +1187,12 @@ where
     }
 }
 
+impl<FD, FS, D, T, B> fmt::Display for BlockListExpand<FD, FS, D, T, B> {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        f.write_str("dense Tensor expansion")
+    }
+}
+
 // TODO: &Txn, not Txn
 type Reductor<FD, FS, D, T> =
     fn(&DenseTensor<FD, FS, D, T, DenseAccessor<FD, FS, D, T>>, T) -> TCBoxTryFuture<Number>;
@@ -1009,12 +1251,13 @@ where
 #[async_trait]
 impl<FD, FS, D, T, B> DenseAccess<FD, FS, D, T> for BlockListReduce<FD, FS, D, T, B>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
-    B: DenseAccess<FD, FS, D, T>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
+    B: DenseAccess<FD, FS, D, T>,
 {
     type Slice = BlockListReduce<FD, FS, D, T, <B as DenseAccess<FD, FS, D, T>>::Slice>;
     type Transpose = BlockListTranspose<FD, FS, D, T, Self>;
@@ -1038,6 +1281,7 @@ where
                     let reductor = self.reductor;
                     let source_bounds = self.rebase.invert_coord(&coord);
                     Box::pin(async move {
+                        debug!("slice {} from {}", source_bounds, source);
                         let slice = source.slice(source_bounds)?;
                         reductor(&slice.accessor().into(), txn.clone()).await
                     })
@@ -1082,12 +1326,13 @@ where
 
 impl<FD, FS, D, T, B> ReadValueAt<D> for BlockListReduce<FD, FS, D, T, B>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
-    B: DenseAccess<FD, FS, D, T>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
+    B: DenseAccess<FD, FS, D, T>,
 {
     type Txn = T;
 
@@ -1104,6 +1349,12 @@ where
     }
 }
 
+impl<FD, FS, D, T, B> fmt::Display for BlockListReduce<FD, FS, D, T, B> {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        f.write_str("dense Tensor reduction")
+    }
+}
+
 #[derive(Clone)]
 pub struct BlockListTranspose<FD, FS, D, T, B> {
     source: B,
@@ -1157,12 +1408,13 @@ where
 #[async_trait]
 impl<FD, FS, D, T, B> DenseAccess<FD, FS, D, T> for BlockListTranspose<FD, FS, D, T, B>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
-    B: DenseAccess<FD, FS, D, T>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
+    B: DenseAccess<FD, FS, D, T>,
 {
     type Slice = <<B as DenseAccess<FD, FS, D, T>>::Slice as DenseAccess<FD, FS, D, T>>::Transpose;
     type Transpose = B::Transpose;
@@ -1230,12 +1482,13 @@ where
 
 impl<FD, FS, D, T, B> ReadValueAt<D> for BlockListTranspose<FD, FS, D, T, B>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
-    B: DenseAccess<FD, FS, D, T>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
+    B: DenseAccess<FD, FS, D, T>,
 {
     type Txn = T;
 
@@ -1251,6 +1504,12 @@ where
     }
 }
 
+impl<FD, FS, D, T, B> fmt::Display for BlockListTranspose<FD, FS, D, T, B> {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        f.write_str("dense Tensor transpose")
+    }
+}
+
 #[derive(Clone)]
 pub struct BlockListSparse<FD, FS, D, T, A> {
     source: A,
@@ -1285,12 +1544,13 @@ where
 #[async_trait]
 impl<FD, FS, D, T, A> DenseAccess<FD, FS, D, T> for BlockListSparse<FD, FS, D, T, A>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
-    A: SparseAccess<FD, FS, D, T>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
+    A: SparseAccess<FD, FS, D, T>,
 {
     type Slice = BlockListSparse<FD, FS, D, T, A::Slice>;
     type Transpose = BlockListSparse<FD, FS, D, T, A::Transpose>;
@@ -1343,12 +1603,13 @@ where
 
 impl<FD, FS, D, T, A> ReadValueAt<D> for BlockListSparse<FD, FS, D, T, A>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
-    A: SparseAccess<FD, FS, D, T>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
+    A: SparseAccess<FD, FS, D, T>,
 {
     type Txn = T;
 
@@ -1366,6 +1627,12 @@ impl<FD, FS, D, T, A> From<A> for BlockListSparse<FD, FS, D, T, A> {
     }
 }
 
+impl<FD, FS, D, T, A> fmt::Display for BlockListSparse<FD, FS, D, T, A> {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        f.write_str("dense representation of a sparse Tensor")
+    }
+}
+
 #[derive(Clone)]
 pub struct BlockListUnary<FD, FS, D, T, B> {
     source: B,
@@ -1427,12 +1694,13 @@ where
 #[async_trait]
 impl<FD, FS, D, T, B> DenseAccess<FD, FS, D, T> for BlockListUnary<FD, FS, D, T, B>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
-    B: DenseAccess<FD, FS, D, T>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
+    B: DenseAccess<FD, FS, D, T>,
 {
     type Slice = BlockListUnary<FD, FS, D, T, B::Slice>;
     type Transpose = BlockListUnary<FD, FS, D, T, B::Transpose>;
@@ -1511,3 +1779,9 @@ where
         })
     }
 }
+
+impl<FD, FS, D, T, B> fmt::Display for BlockListUnary<FD, FS, D, T, B> {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        f.write_str("dense Tensor unary op")
+    }
+}
