@@ -1,7 +1,7 @@
-use std::convert::TryFrom;
 use std::fmt;
 use std::iter::{self, FromIterator};
 use std::marker::PhantomData;
+use std::ops::Deref;
 
 use afarray::{Array, ArrayExt, ArrayInstance, CoordBlocks, Coords, Offsets};
 use arrayfire as af;
@@ -10,6 +10,7 @@ use destream::de;
 use futures::stream::{self, Stream, StreamExt, TryStreamExt};
 use futures::{future, try_join, TryFutureExt};
 use log::debug;
+use safecast::AsType;
 use strided::Stride;
 
 use tc_btree::Node;
@@ -24,9 +25,10 @@ use crate::transform;
 use crate::{coord_bounds, Bounds, Coord, Schema, Shape, TensorAccess, TensorType};
 
 use super::access::BlockListTranspose;
-use super::{DenseAccess, DenseAccessor, DenseWrite, PER_BLOCK};
+use super::{DenseAccess, DenseAccessor, DenseWrite, MEBIBYTE, PER_BLOCK};
 
-const MEBIBYTE: usize = 1_048_576;
+/// The size of a dense tensor block on disk, in bytes (1 mebibyte + 5 bytes overhead).
+const BLOCK_SIZE: usize = MEBIBYTE + 5;
 
 /// A wrapper around a `DenseTensor` [`File`]
 #[derive(Clone)]
@@ -61,7 +63,7 @@ where
     pub async fn constant(file: FD, txn_id: TxnId, shape: Shape, value: Number) -> TCResult<Self> {
         debug!("BlockListFile::constant {} {}", shape, value);
 
-        if !file.is_empty(&txn_id).await? {
+        if !file.is_empty(txn_id).await? {
             return Err(TCError::unsupported(
                 "cannot create new tensor: file is not empty",
             ));
@@ -109,7 +111,8 @@ where
             .map(|(i, r)| r.map(|block| (BlockId::from(i), block)))
             .map_ok(|(id, block)| {
                 let len = block.len() as u64;
-                file.create_block(txn_id, id, block).map_ok(move |_| len)
+                file.create_block(txn_id, id, block, BLOCK_SIZE)
+                    .map_ok(move |_| len)
             })
             .try_buffer_unordered(num_cpus::get())
             .try_fold(0u64, |block_len, size| future::ready(Ok(size + block_len)))
@@ -151,7 +154,9 @@ where
             size += chunk.len() as u64;
             let block_id = BlockId::from(i);
             let block = Array::from(chunk).cast_into(dtype);
-            file.create_block(txn_id, block_id, block).await?;
+            file.create_block(txn_id, block_id, block, BLOCK_SIZE)
+                .await?;
+
             i += 1;
         }
 
@@ -211,6 +216,7 @@ where
         } else if num_blocks == 1 {
             let block_id = BlockId::from(0u64);
             let mut block = self.file.write_block(txn_id, block_id).await?;
+
             block.sort(true)?;
             return Ok(());
         }
@@ -220,10 +226,12 @@ where
             let block_id = BlockId::from(block_id);
 
             let left = self.file.write_block(txn_id, block_id);
+
             let right = self.file.write_block(txn_id, next_block_id);
+
             let (mut left, mut right) = try_join!(left, right)?;
 
-            let mut block = Array::concatenate(&left, &right)?;
+            let mut block = Array::concatenate(&left, &right);
             block.sort(true)?;
 
             let (left_sorted, right_sorted) = block.split(PER_BLOCK)?;
@@ -272,6 +280,7 @@ where
             .map(|(block_id, r)| r.map(|array| (block_id, array)))
             .map_ok(|(block_id, array)| async {
                 let mut block = self.file.write_block(txn_id, block_id).await?;
+
                 *block = array;
                 Ok(())
             })
@@ -302,10 +311,11 @@ impl<FD: Send, FS: Send, D: Send, T: Send> TensorAccess for BlockListFile<FD, FS
 #[async_trait]
 impl<FD, FS, D, T> DenseAccess<FD, FS, D, T> for BlockListFile<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
 {
     type Slice = BlockListFileSlice<FD, FS, D, T>;
@@ -357,8 +367,8 @@ where
         stream::iter(block_ids)
             .map(|block_id| {
                 let af_block_id = af::Array::new(&[block_id], af::Dim4::new(&[1, 1, 1, 1]));
-                let mask = ArrayExt::from(af::eq(block_offsets.af(), &af_block_id, true));
-                let indices = (&offsets % &per_block).af() * mask.af();
+                let mask = ArrayExt::from(af::eq(block_offsets.deref(), &af_block_id, true));
+                let indices = (&offsets % &per_block).deref() * mask.deref();
                 (block_id, mask.into(), indices.into())
             })
             .map(|(block_id, mask, indices)| {
@@ -377,10 +387,11 @@ where
 #[async_trait]
 impl<FD, FS, D, T> DenseWrite<FD, FS, D, T> for BlockListFile<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
 {
     async fn write<B: DenseAccess<FD, FS, D, T>>(
@@ -431,15 +442,16 @@ where
 
                 async move {
                     let indices: ArrayExt<u64> =
-                        af::modulo(offsets.af(), &af_per_block, true).into();
+                        af::modulo(offsets.deref(), &af_per_block, true).into();
 
-                    let block_offsets = af::div(offsets.af(), &af_per_block, true);
+                    let block_offsets = af::div(offsets.deref(), &af_per_block, true);
                     let block_ids = ArrayExt::from(af::set_unique(&block_offsets, true)).to_vec();
 
                     let mut start = 0;
                     for block_id in block_ids.into_iter() {
                         let af_block_id = ArrayExt::from(&[block_id][..]);
-                        let (len, _) = af::sum_all(&af::eq(&block_offsets, af_block_id.af(), true));
+                        let (len, _) =
+                            af::sum_all(&af::eq(&block_offsets, af_block_id.deref(), true));
                         let end = start + len as usize;
                         let indices = indices.slice(start, end);
                         let array = array.slice(start, end).map_err(TCError::from)?;
@@ -506,10 +518,11 @@ where
 
 impl<FD, FS, D, T> ReadValueAt<D> for BlockListFile<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
 {
     type Txn = T;
@@ -526,6 +539,7 @@ where
 
             let block_id = BlockId::from(offset / PER_BLOCK as u64);
             let block = self.file.read_block(*txn.id(), block_id).await?;
+
             let value = block.get_value((offset % PER_BLOCK as u64) as usize);
 
             Ok((coord, value))
@@ -617,6 +631,12 @@ where
     }
 }
 
+impl<FD, FS, D, T> fmt::Display for BlockListFile<FD, FS, D, T> {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        f.write_str("dense Tensor file")
+    }
+}
+
 #[async_trait]
 impl<FD, FS, D, T> de::FromStream for BlockListFile<FD, FS, D, T>
 where
@@ -707,14 +727,14 @@ impl<'a, F: File<Array>> BlockListVisitor<'a, F> {
         &self,
         block_id: u64,
         block: ArrayExt<T>,
-    ) -> Result<<F as File<Array>>::Block, E>
+    ) -> Result<F::Write, E>
     where
         Array: From<ArrayExt<T>>,
     {
         debug!("BlockListVisitor::create_block {}", block_id);
 
         self.file
-            .create_block(self.txn_id, block_id.into(), block.into())
+            .create_block(self.txn_id, block_id.into(), block.into(), BLOCK_SIZE)
             .map_err(de::Error::custom)
             .await
     }
@@ -732,6 +752,11 @@ impl<'a, F: File<Array>> BlockListVisitor<'a, F> {
     where
         Array: From<ArrayExt<T>>,
     {
+        debug!(
+            "BlockListVisitor decoding array of type {}",
+            std::any::type_name::<T>()
+        );
+
         let mut buf = vec![T::default(); BUF_SIZE];
         let mut size = 0u64;
         let mut block_id = 0u64;
@@ -873,6 +898,7 @@ impl<'a, F: File<Array>> ComplexBlockListVisitor<'a, F> {
                 let im = ArrayExt::<T>::from_iter(im.iter().cloned());
                 let block = ArrayExt::from((re, im));
                 self.visitor.create_block(block_id, block).await?;
+
                 size += block_size as u64;
                 block_id += 1;
             }
@@ -955,10 +981,11 @@ where
 #[async_trait]
 impl<FD, FS, D, T> DenseAccess<FD, FS, D, T> for BlockListFileSlice<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
 {
     type Slice = Self;
@@ -994,6 +1021,7 @@ where
                         block_offsets(&af_indices, &af_offsets, start, block_id);
 
                     let block = file_clone.read_block(txn_id, block_id.into()).await?;
+
                     values.extend(block.get(&block_offsets.into()).to_vec());
                     start = new_start;
                 }
@@ -1024,10 +1052,11 @@ where
 
 impl<FD, FS, D, T> ReadValueAt<D> for BlockListFileSlice<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
 {
     type Txn = T;
@@ -1042,6 +1071,12 @@ where
     }
 }
 
+impl<FD, FS, D, T> fmt::Display for BlockListFileSlice<FD, FS, D, T> {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        f.write_str("dense Tensor slice")
+    }
+}
+
 #[inline]
 fn div_ceil(l: u64, r: u64) -> u64 {
     if l % r == 0 {
@@ -1060,7 +1095,7 @@ fn block_offsets(
     assert_eq!(indices.len(), offsets.len());
 
     let num_to_update = af::sum_all(&af::eq(
-        indices.af(),
+        indices.deref(),
         &af::constant(block_id, af::Dim4::new(&[1, 1, 1, 1])),
         true,
     ))
@@ -1080,7 +1115,7 @@ fn coord_block(coords: Coords, shape: &[u64]) -> (Vec<u64>, ArrayExt<u64>, Offse
     let af_per_block = af::constant(PER_BLOCK as u64, af::Dim4::new(&[1, 1, 1, 1]));
 
     let offsets = coords.to_offsets(shape);
-    let block_offsets = ArrayExt::from(af::div(offsets.af(), &af_per_block, true));
+    let block_offsets = ArrayExt::from(af::div(offsets.deref(), &af_per_block, true));
     let block_ids = block_offsets.unique(true);
     (block_ids.to_vec(), block_offsets, offsets)
 }
