@@ -1,4 +1,3 @@
-use std::convert::TryFrom;
 use std::fmt;
 use std::marker::PhantomData;
 use std::ops::{Add, Div, Mul, Sub};
@@ -8,33 +7,37 @@ use arrayfire as af;
 use async_trait::async_trait;
 use destream::{de, en, EncodeSeq};
 use futures::future::{self, TryFutureExt};
-use futures::stream::{Stream, TryStreamExt};
+use futures::stream::{Stream, StreamExt, TryStreamExt};
 use log::{debug, warn};
+use safecast::{AsType, CastFrom};
 
 use tc_btree::Node;
 use tc_error::*;
 use tc_transact::fs::{CopyFrom, Dir, File, Hash, Persist, Restore};
 use tc_transact::{IntoView, Transact, Transaction, TxnId};
-use tc_value::{Number, NumberClass, NumberInstance, NumberType};
+use tc_value::{FloatType, Number, NumberClass, NumberInstance, NumberType};
 use tcgeneric::{TCBoxTryFuture, TCBoxTryStream};
 
 use super::sparse::{DenseToSparse, SparseTensor};
 use super::stream::{Read, ReadValueAt};
 use super::{
-    Bounds, Coord, Phantom, Schema, Shape, Tensor, TensorAccess, TensorBoolean, TensorCompare,
-    TensorDualIO, TensorIO, TensorInstance, TensorMath, TensorReduce, TensorTransform, TensorType,
-    TensorUnary,
+    Bounds, Coord, Phantom, Schema, Shape, Tensor, TensorAccess, TensorBoolean, TensorBooleanConst,
+    TensorCompare, TensorCompareConst, TensorDualIO, TensorIO, TensorInstance, TensorMath,
+    TensorMathConst, TensorReduce, TensorTransform, TensorType, TensorUnary, ERR_COMPLEX_EXPONENT,
 };
 
 use access::*;
 pub use access::{BlockListSparse, DenseAccess, DenseAccessor, DenseWrite};
 pub use file::BlockListFile;
-use futures::StreamExt;
 
+#[allow(unused)]
 mod access;
 mod file;
 mod stream;
 
+/// The number of bytes in one mebibyte.s
+const MEBIBYTE: usize = 1_048_576;
+
 /// The number of elements per dense tensor block, equal to (1 mebibyte / 64 bits).
 pub const PER_BLOCK: usize = 131_072;
 
@@ -54,10 +57,11 @@ impl<FD, FS, D, T, B> DenseTensor<FD, FS, D, T, B> {
 
 impl<FD, FS, D, T, B> DenseTensor<FD, FS, D, T, B>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     B: DenseAccess<FD, FS, D, T>,
 {
     fn combine<OT: DenseAccess<FD, FS, D, T>>(
@@ -89,10 +93,11 @@ where
 
 impl<FD, FS, D, T> DenseTensor<FD, FS, D, T, BlockListFile<FD, FS, D, T>>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
 {
     /// Create a new `DenseTensor` with the given [`Schema`].
     pub async fn create(file: FD, schema: Schema, txn_id: TxnId) -> TCResult<Self> {
@@ -170,10 +175,11 @@ impl<FD, FS, D, T, B> TensorInstance for DenseTensor<FD, FS, D, T, B> {
 impl<FD, FS, D, T, B, O> TensorBoolean<DenseTensor<FD, FS, D, T, O>>
     for DenseTensor<FD, FS, D, T, B>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     B: DenseAccess<FD, FS, D, T>,
     O: DenseAccess<FD, FS, D, T>,
     D::FileClass: From<TensorType>,
@@ -196,10 +202,11 @@ where
 
 impl<FD, FS, D, T, B> TensorBoolean<Tensor<FD, FS, D, T>> for DenseTensor<FD, FS, D, T, B>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     B: DenseAccess<FD, FS, D, T>,
     D::FileClass: From<TensorType>,
 {
@@ -228,13 +235,52 @@ where
     }
 }
 
+impl<FD, FS, D, T, B> TensorBooleanConst for DenseTensor<FD, FS, D, T, B>
+where
+    D: Dir,
+    T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    B: DenseAccess<FD, FS, D, T>,
+{
+    type Combine = DenseTensor<FD, FS, D, T, BlockListConst<FD, FS, D, T, B>>;
+
+    fn and_const(self, other: Number) -> TCResult<Self::Combine> {
+        fn array_and(l: Array, r: Number) -> Array {
+            l.and_const(r)
+        }
+
+        let blocks = BlockListConst::new(self.blocks, other, array_and, Number::and);
+        Ok(blocks.into())
+    }
+
+    fn or_const(self, other: Number) -> TCResult<Self::Combine> {
+        fn array_or(l: Array, r: Number) -> Array {
+            l.or_const(r)
+        }
+
+        let blocks = BlockListConst::new(self.blocks, other, array_or, Number::or);
+        Ok(blocks.into())
+    }
+
+    fn xor_const(self, other: Number) -> TCResult<Self::Combine> {
+        fn array_xor(l: Array, r: Number) -> Array {
+            l.xor_const(r)
+        }
+
+        let blocks = BlockListConst::new(self.blocks, other, array_xor, Number::xor);
+        Ok(blocks.into())
+    }
+}
+
 impl<FD, FS, D, T, B, O> TensorCompare<DenseTensor<FD, FS, D, T, O>>
     for DenseTensor<FD, FS, D, T, B>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     B: DenseAccess<FD, FS, D, T>,
     O: DenseAccess<FD, FS, D, T>,
     D::FileClass: From<TensorType>,
@@ -293,10 +339,11 @@ where
 
 impl<FD, FS, D, T, B> TensorCompare<Tensor<FD, FS, D, T>> for DenseTensor<FD, FS, D, T, B>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     B: DenseAccess<FD, FS, D, T>,
     D::FileClass: From<TensorType>,
 {
@@ -346,13 +393,97 @@ where
     }
 }
 
+impl<FD, FS, D, T, B> TensorCompareConst for DenseTensor<FD, FS, D, T, B>
+where
+    D: Dir,
+    T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    B: DenseAccess<FD, FS, D, T>,
+{
+    type Compare = DenseTensor<FD, FS, D, T, BlockListConst<FD, FS, D, T, B>>;
+
+    fn eq_const(self, other: Number) -> TCResult<Self::Compare> {
+        fn eq_array(l: Array, r: Number) -> Array {
+            l.eq_const(r)
+        }
+
+        fn eq_number(l: Number, r: Number) -> Number {
+            (l == r).into()
+        }
+
+        Ok(BlockListConst::new(self.blocks, other, eq_array, eq_number).into())
+    }
+
+    fn gt_const(self, other: Number) -> TCResult<Self::Compare> {
+        fn gt_array(l: Array, r: Number) -> Array {
+            l.gt_const(r)
+        }
+
+        fn gt_number(l: Number, r: Number) -> Number {
+            (l > r).into()
+        }
+
+        Ok(BlockListConst::new(self.blocks, other, gt_array, gt_number).into())
+    }
+
+    fn gte_const(self, other: Number) -> TCResult<Self::Compare> {
+        fn gte_array(l: Array, r: Number) -> Array {
+            l.gte_const(r)
+        }
+
+        fn gte_number(l: Number, r: Number) -> Number {
+            (l >= r).into()
+        }
+
+        Ok(BlockListConst::new(self.blocks, other, gte_array, gte_number).into())
+    }
+
+    fn lt_const(self, other: Number) -> TCResult<Self::Compare> {
+        fn lt_array(l: Array, r: Number) -> Array {
+            l.lt_const(r)
+        }
+
+        fn lt_number(l: Number, r: Number) -> Number {
+            (l < r).into()
+        }
+
+        Ok(BlockListConst::new(self.blocks, other, lt_array, lt_number).into())
+    }
+
+    fn lte_const(self, other: Number) -> TCResult<Self::Compare> {
+        fn lte_array(l: Array, r: Number) -> Array {
+            l.lte_const(r)
+        }
+
+        fn lte_number(l: Number, r: Number) -> Number {
+            (l <= r).into()
+        }
+
+        Ok(BlockListConst::new(self.blocks, other, lte_array, lte_number).into())
+    }
+
+    fn ne_const(self, other: Number) -> TCResult<Self::Compare> {
+        fn ne_array(l: Array, r: Number) -> Array {
+            l.ne_const(r)
+        }
+
+        fn ne_number(l: Number, r: Number) -> Number {
+            (l != r).into()
+        }
+
+        Ok(BlockListConst::new(self.blocks, other, ne_array, ne_number).into())
+    }
+}
+
 #[async_trait]
 impl<FD, FS, D, T, B> TensorIO<D> for DenseTensor<FD, FS, D, T, B>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     B: DenseWrite<FD, FS, D, T>,
     D::FileClass: From<TensorType>,
 {
@@ -382,10 +513,11 @@ where
 impl<FD, FS, D, T, B, O> TensorDualIO<D, DenseTensor<FD, FS, D, T, O>>
     for DenseTensor<FD, FS, D, T, B>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     B: DenseWrite<FD, FS, D, T>,
     O: DenseAccess<FD, FS, D, T>,
     D::FileClass: From<TensorType>,
@@ -406,10 +538,11 @@ where
 #[async_trait]
 impl<FD, FS, D, T, B> TensorDualIO<D, Tensor<FD, FS, D, T>> for DenseTensor<FD, FS, D, T, B>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     B: DenseWrite<FD, FS, D, T>,
     D::FileClass: From<TensorType>,
 {
@@ -435,10 +568,11 @@ where
 impl<FD, FS, D, T, B, O> TensorMath<D, DenseTensor<FD, FS, D, T, O>>
     for DenseTensor<FD, FS, D, T, B>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     B: DenseAccess<FD, FS, D, T>,
     O: DenseAccess<FD, FS, D, T>,
     D::FileClass: From<TensorType>,
@@ -480,9 +614,21 @@ where
         self.combine(other, mul_array, Mul::mul, dtype)
     }
 
-    fn sub(self, other: DenseTensor<FD, FS, D, T, O>) -> TCResult<Self::Combine> {
-        debug!("subtract {} from {}", other, self);
+    fn pow(self, other: DenseTensor<FD, FS, D, T, O>) -> TCResult<Self::Combine> {
+        if !other.dtype().is_real() {
+            return Err(TCError::unsupported(ERR_COMPLEX_EXPONENT));
+        }
 
+        fn pow_array(l: &Array, r: &Array) -> Array {
+            debug_assert_eq!(l.len(), r.len());
+            l.pow(r)
+        }
+
+        let dtype = Ord::max(self.dtype(), other.dtype());
+        self.combine(other, pow_array, Number::pow, dtype)
+    }
+
+    fn sub(self, other: DenseTensor<FD, FS, D, T, O>) -> TCResult<Self::Combine> {
         fn sub_array(l: &Array, r: &Array) -> Array {
             debug_assert_eq!(l.len(), r.len());
             l - r
@@ -495,10 +641,11 @@ where
 
 impl<FD, FS, D, T, B> TensorMath<D, Tensor<FD, FS, D, T>> for DenseTensor<FD, FS, D, T, B>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     B: DenseAccess<FD, FS, D, T>,
     D::FileClass: From<TensorType>,
 {
@@ -526,6 +673,13 @@ where
         }
     }
 
+    fn pow(self, other: Tensor<FD, FS, D, T>) -> TCResult<Self::Combine> {
+        match other {
+            Tensor::Dense(dense) => self.pow(dense).map(Tensor::from),
+            Tensor::Sparse(sparse) => sparse.pow(self.into_sparse()).map(Tensor::from),
+        }
+    }
+
     fn sub(self, other: Tensor<FD, FS, D, T>) -> TCResult<Self::Combine> {
         match other {
             Tensor::Dense(dense) => self.sub(dense).map(Tensor::from),
@@ -534,12 +688,68 @@ where
     }
 }
 
-impl<FD, FS, D, T, B> ReadValueAt<D> for DenseTensor<FD, FS, D, T, B>
+impl<FD, FS, D, T, B> TensorMathConst for DenseTensor<FD, FS, D, T, B>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    B: DenseAccess<FD, FS, D, T>,
+{
+    type Combine = DenseTensor<FD, FS, D, T, BlockListConst<FD, FS, D, T, B>>;
+
+    fn add_const(self, other: Number) -> TCResult<Self::Combine> {
+        fn add_array(l: Array, r: Number) -> Array {
+            &l + r
+        }
+
+        Ok(BlockListConst::new(self.blocks, other, add_array, Number::add).into())
+    }
+
+    fn div_const(self, other: Number) -> TCResult<Self::Combine> {
+        fn div_array(l: Array, r: Number) -> Array {
+            &l / r
+        }
+
+        Ok(BlockListConst::new(self.blocks, other, div_array, Number::div).into())
+    }
+
+    fn mul_const(self, other: Number) -> TCResult<Self::Combine> {
+        fn mul_array(l: Array, r: Number) -> Array {
+            &l * r
+        }
+
+        Ok(BlockListConst::new(self.blocks, other, mul_array, Number::mul).into())
+    }
+
+    fn pow_const(self, other: Number) -> TCResult<Self::Combine> {
+        if !other.class().is_real() {
+            return Err(TCError::unsupported(ERR_COMPLEX_EXPONENT));
+        }
+
+        fn pow_array(l: Array, r: Number) -> Array {
+            l.pow_const(r)
+        }
+
+        Ok(BlockListConst::new(self.blocks, other, pow_array, Number::pow).into())
+    }
+
+    fn sub_const(self, other: Number) -> TCResult<Self::Combine> {
+        fn sub_array(l: Array, r: Number) -> Array {
+            &l - r
+        }
+
+        Ok(BlockListConst::new(self.blocks, other, sub_array, Number::sub).into())
+    }
+}
+
+impl<FD, FS, D, T, B> ReadValueAt<D> for DenseTensor<FD, FS, D, T, B>
+where
+    D: Dir,
+    T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     B: DenseAccess<FD, FS, D, T>,
     D::FileClass: From<TensorType>,
 {
@@ -552,10 +762,11 @@ where
 
 impl<FD, FS, D, T, B> TensorReduce<D> for DenseTensor<FD, FS, D, T, B>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     B: DenseAccess<FD, FS, D, T>,
     D::FileClass: From<TensorType>,
 {
@@ -605,10 +816,11 @@ where
 
 impl<FD, FS, D, T, B> TensorTransform for DenseTensor<FD, FS, D, T, B>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     B: DenseAccess<FD, FS, D, T>,
     D::FileClass: From<TensorType>,
 {
@@ -647,10 +859,11 @@ where
 #[async_trait]
 impl<FD, FS, D, T, B> TensorUnary<D> for DenseTensor<FD, FS, D, T, B>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     B: DenseAccess<FD, FS, D, T>,
     D::FileClass: From<TensorType>,
 {
@@ -668,6 +881,22 @@ where
         Ok(DenseTensor::from(blocks))
     }
 
+    fn exp(&self) -> TCResult<Self::Unary> {
+        fn exp(n: Number) -> Number {
+            let n = f64::cast_from(n);
+            n.exp().into()
+        }
+
+        let blocks = BlockListUnary::new(
+            self.blocks.clone(),
+            Array::exp,
+            exp,
+            NumberType::Float(FloatType::F64),
+        );
+
+        Ok(DenseTensor::from(blocks))
+    }
+
     async fn all(self, txn: T) -> TCResult<bool> {
         let mut blocks = self.blocks.block_stream(txn).await?;
 
@@ -706,10 +935,12 @@ where
 #[async_trait]
 impl<FD, FS, D, T> Transact for DenseTensor<FD, FS, D, T, BlockListFile<FD, FS, D, T>>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError> + Transact,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array> + Transact,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
+    T: Transaction<D>,
     D::FileClass: From<TensorType>,
 {
     async fn commit(&self, txn_id: &TxnId) {
@@ -724,10 +955,11 @@ where
 #[async_trait]
 impl<'en, FD, FS, D, T, B> Hash<'en, D> for DenseTensor<FD, FS, D, T, B>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     B: DenseAccess<FD, FS, D, T>,
     D::FileClass: From<TensorType>,
 {
@@ -743,10 +975,11 @@ where
 impl<FD, FS, D, T, B> CopyFrom<D, DenseTensor<FD, FS, D, T, B>>
     for DenseTensor<FD, FS, D, T, BlockListFile<FD, FS, D, T>>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     B: DenseAccess<FD, FS, D, T>,
     D::FileClass: From<TensorType>,
 {
@@ -764,10 +997,12 @@ where
 #[async_trait]
 impl<FD, FS, D, T> Persist<D> for DenseTensor<FD, FS, D, T, BlockListFile<FD, FS, D, T>>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
+    T: Transaction<D>,
     D::FileClass: From<TensorType>,
 {
     type Schema = Schema;
@@ -788,10 +1023,12 @@ where
 #[async_trait]
 impl<FD, FS, D, T> Restore<D> for DenseTensor<FD, FS, D, T, BlockListFile<FD, FS, D, T>>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
+    T: Transaction<D>,
     D::FileClass: From<TensorType>,
 {
     async fn restore(&self, backup: &Self, txn_id: TxnId) -> TCResult<()> {
@@ -811,10 +1048,12 @@ impl<FD, FS, D, T, B> From<B> for DenseTensor<FD, FS, D, T, B> {
 #[async_trait]
 impl<FD, FS, D, T> de::FromStream for DenseTensor<FD, FS, D, T, BlockListFile<FD, FS, D, T>>
 where
-    FD: File<Array> + TryFrom<<D as Dir>::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
+    T: Transaction<D>,
     D::FileClass: From<TensorType> + Send,
 {
     type Context = T;
@@ -823,7 +1062,7 @@ where
         let txn_id = *txn.id();
         let file = txn
             .context()
-            .create_file_tmp(txn_id, TensorType::Dense)
+            .create_file_unique(txn_id, TensorType::Dense)
             .map_err(de::Error::custom)
             .await?;
 
@@ -869,10 +1108,12 @@ impl<FD, FS, D, T> DenseTensorVisitor<FD, FS, D, T> {
 #[async_trait]
 impl<FD, FS, D, T> de::Visitor for DenseTensorVisitor<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
+    T: Transaction<D>,
     D::FileClass: From<TensorType>,
 {
     type Value = DenseTensor<FD, FS, D, T, BlockListFile<FD, FS, D, T>>;
@@ -882,14 +1123,16 @@ where
     }
 
     async fn visit_seq<A: de::SeqAccess>(self, mut seq: A) -> Result<Self::Value, A::Error> {
+        debug!("deserialize DenseTensor");
+
         let schema = seq.next_element(()).await?;
         let schema = schema.ok_or_else(|| de::Error::invalid_length(0, "a tensor schema"))?;
+        debug!("DenseTensor schema is {}", schema);
 
         let cxt = (self.txn_id, self.file, schema);
-        let blocks = seq
-            .next_element::<BlockListFile<FD, FS, D, T>>(cxt)
-            .await?
-            .ok_or_else(|| de::Error::invalid_length(1, "dense tensor data"))?;
+        let blocks = seq.next_element::<BlockListFile<FD, FS, D, T>>(cxt).await?;
+
+        let blocks = blocks.ok_or_else(|| de::Error::invalid_length(1, "dense tensor data"))?;
 
         Ok(DenseTensor::from(blocks))
     }
@@ -898,10 +1141,11 @@ where
 #[async_trait]
 impl<'en, FD, FS, D, T, B> IntoView<'en, D> for DenseTensor<FD, FS, D, T, B>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     B: DenseAccess<FD, FS, D, T>,
     D::FileClass: From<TensorType>,
 {
@@ -953,9 +1197,16 @@ impl<'en> en::IntoStream<'en> for BlockStreamView<'en> {
             // an error can't be encoded within an array
             // so in case of a read error, let the receiver figure out that the tensor
             // doesn't have enough elements
-            blocks
-                .take_while(|r| future::ready(r.is_ok()))
-                .map(|r| r.expect("tensor block").type_cast().to_vec())
+
+            #[cfg(not(debug_assertions))]
+            {
+                blocks
+                    .take_while(|r| future::ready(r.is_ok()))
+                    .map(|r| r.expect("tensor block").type_cast().to_vec())
+            }
+
+            #[cfg(debug_assertions)]
+            blocks.map(|r| r.expect("tensor block").type_cast().to_vec())
         }
 
         match self.dtype {
