@@ -1,9 +1,7 @@
 /// A [`Tensor`], an n-dimensional array of [`Number`]s which supports basic math and logic
-use std::convert::TryFrom;
 use std::fmt;
 use std::marker::PhantomData;
 
-use afarray::Array;
 use async_trait::async_trait;
 use destream::{de, en};
 use futures::TryFutureExt;
@@ -20,6 +18,7 @@ use tcgeneric::{
     TCPathBuf, Tuple,
 };
 
+pub use afarray::{print_af_info, Array};
 pub use bounds::{AxisBounds, Bounds, Shape};
 pub use dense::{BlockListFile, DenseAccess, DenseAccessor, DenseTensor, DenseWrite};
 pub use einsum::einsum;
@@ -32,6 +31,7 @@ mod sparse;
 mod stream;
 mod transform;
 
+const ERR_COMPLEX_EXPONENT: &str = "raising to a complex power is not supported";
 const ERR_INF: &str = "Tensor combination resulted in an infinite value";
 const ERR_NAN: &str = "Tensor combination resulted in a non-numeric value";
 
@@ -86,6 +86,7 @@ impl de::FromStream for Schema {
     type Context = ();
 
     async fn from_stream<D: de::Decoder>(cxt: (), decoder: &mut D) -> Result<Self, D::Error> {
+        debug!("deserialize Tensor schema");
         let schema = Value::from_stream(cxt, decoder).await?;
         Self::try_cast_from(schema, |v| de::Error::invalid_value(v, "a Tensor schema"))
     }
@@ -156,6 +157,21 @@ pub trait TensorBoolean<O> {
     fn xor(self, other: O) -> TCResult<Self::Combine>;
 }
 
+/// [`Tensor`] boolean operations in relation to a constant.
+pub trait TensorBooleanConst {
+    /// The result type of a boolean operation.
+    type Combine: TensorInstance;
+
+    /// Logical and
+    fn and_const(self, other: Number) -> TCResult<Self::Combine>;
+
+    /// Logical or
+    fn or_const(self, other: Number) -> TCResult<Self::Combine>;
+
+    /// Logical xor
+    fn xor_const(self, other: Number) -> TCResult<Self::Combine>;
+}
+
 /// Tensor comparison operations
 pub trait TensorCompare<O> {
     /// The result of a comparison operation
@@ -183,6 +199,30 @@ pub trait TensorCompare<O> {
     fn ne(self, other: O) -> TCResult<Self::Compare>;
 }
 
+/// Tensor-constant comparison operations
+pub trait TensorCompareConst {
+    /// The result of a comparison operation
+    type Compare: TensorInstance;
+
+    /// Element-wise equality
+    fn eq_const(self, other: Number) -> TCResult<Self::Compare>;
+
+    /// Element-wise greater-than
+    fn gt_const(self, other: Number) -> TCResult<Self::Compare>;
+
+    /// Element-wise greater-or-equal
+    fn gte_const(self, other: Number) -> TCResult<Self::Compare>;
+
+    /// Element-wise less-than
+    fn lt_const(self, other: Number) -> TCResult<Self::Compare>;
+
+    /// Element-wise less-or-equal
+    fn lte_const(self, other: Number) -> TCResult<Self::Compare>;
+
+    /// Element-wise not-equal
+    fn ne_const(self, other: Number) -> TCResult<Self::Compare>;
+}
+
 /// [`Tensor`] I/O operations
 #[async_trait]
 pub trait TensorIO<D: Dir> {
@@ -226,10 +266,34 @@ pub trait TensorMath<D: Dir, O> {
     /// Multiply two tensors together.
     fn mul(self, other: O) -> TCResult<Self::LeftCombine>;
 
+    /// Raise `self` to the power `other`.
+    fn pow(self, other: O) -> TCResult<Self::LeftCombine>;
+
     /// Subtract `other` from `self`.
     fn sub(self, other: O) -> TCResult<Self::Combine>;
 }
 
+/// [`Tensor`] constant math operations
+pub trait TensorMathConst {
+    /// The result type of a math operation
+    type Combine: TensorInstance;
+
+    /// Add a constant to this tensor
+    fn add_const(self, other: Number) -> TCResult<Self::Combine>;
+
+    /// Divide `self` by `other`.
+    fn div_const(self, other: Number) -> TCResult<Self::Combine>;
+
+    /// Multiply `self` by `other`.
+    fn mul_const(self, other: Number) -> TCResult<Self::Combine>;
+
+    /// Raise `self` to the power `other`.
+    fn pow_const(self, other: Number) -> TCResult<Self::Combine>;
+
+    /// Subtract `other` from `self`.
+    fn sub_const(self, other: Number) -> TCResult<Self::Combine>;
+}
+
 /// [`Tensor`] reduction operations
 pub trait TensorReduce<D: Dir> {
     /// The type of [`Transaction`] to expect
@@ -297,6 +361,9 @@ pub trait TensorUnary<D: Dir> {
     /// Element-wise absolute value
     fn abs(&self) -> TCResult<Self::Unary>;
 
+    /// Raise `e` to the power of `self`
+    fn exp(&self) -> TCResult<Self::Unary>;
+
     /// Return `true` if all elements in this [`Tensor`] are nonzero.
     async fn all(self, txn: Self::Txn) -> TCResult<bool>;
 
@@ -352,10 +419,11 @@ pub enum Tensor<FD, FS, D, T> {
 
 impl<FD, FS, D, T> Tensor<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
 {
     pub fn schema(&self) -> Schema {
@@ -379,10 +447,11 @@ impl<FD: File<Array>, FS: File<Node>, D: Dir, T: Transaction<D>> Instance for Te
 
 impl<FD, FS, D, T> TensorAccess for Tensor<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
 {
     fn dtype(&self) -> NumberType {
@@ -416,10 +485,11 @@ where
 
 impl<FD, FS, D, T> TensorInstance for Tensor<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
 {
     type Dense = Self;
@@ -442,10 +512,11 @@ where
 
 impl<FD, FS, D, T> TensorBoolean<Self> for Tensor<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
 {
     type Combine = Self;
@@ -473,12 +544,46 @@ where
     }
 }
 
-impl<FD, FS, D, T> TensorCompare<Self> for Tensor<FD, FS, D, T>
+impl<FD, FS, D, T> TensorBooleanConst for Tensor<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
+    D::FileClass: From<TensorType>,
+{
+    type Combine = Self;
+
+    fn and_const(self, other: Number) -> TCResult<Self::Combine> {
+        match self {
+            Self::Dense(dense) => dense.and_const(other).map(Self::from),
+            Self::Sparse(sparse) => sparse.and_const(other).map(Self::from),
+        }
+    }
+
+    fn or_const(self, other: Number) -> TCResult<Self::Combine> {
+        match self {
+            Self::Dense(dense) => dense.or_const(other).map(Self::from),
+            Self::Sparse(sparse) => sparse.or_const(other).map(Self::from),
+        }
+    }
+
+    fn xor_const(self, other: Number) -> TCResult<Self::Combine> {
+        match self {
+            Self::Dense(dense) => dense.xor_const(other).map(Self::from),
+            Self::Sparse(sparse) => sparse.xor_const(other).map(Self::from),
+        }
+    }
+}
+
+impl<FD, FS, D, T> TensorCompare<Self> for Tensor<FD, FS, D, T>
+where
+    D: Dir,
+    T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
 {
     type Compare = Self;
@@ -527,13 +632,68 @@ where
     }
 }
 
+impl<FD, FS, D, T> TensorCompareConst for Tensor<FD, FS, D, T>
+where
+    D: Dir,
+    T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
+    D::FileClass: From<TensorType>,
+{
+    type Compare = Self;
+
+    fn eq_const(self, other: Number) -> TCResult<Self::Compare> {
+        match self {
+            Self::Dense(dense) => dense.eq_const(other).map(Self::from),
+            Self::Sparse(sparse) => sparse.eq_const(other).map(Self::from),
+        }
+    }
+
+    fn gt_const(self, other: Number) -> TCResult<Self::Compare> {
+        match self {
+            Self::Dense(dense) => dense.gt_const(other).map(Self::from),
+            Self::Sparse(sparse) => sparse.gt_const(other).map(Self::from),
+        }
+    }
+
+    fn gte_const(self, other: Number) -> TCResult<Self::Compare> {
+        match self {
+            Self::Dense(dense) => dense.gte_const(other).map(Self::from),
+            Self::Sparse(sparse) => sparse.gte_const(other).map(Self::from),
+        }
+    }
+
+    fn lt_const(self, other: Number) -> TCResult<Self::Compare> {
+        match self {
+            Self::Dense(dense) => dense.lt_const(other).map(Self::from),
+            Self::Sparse(sparse) => sparse.lt_const(other).map(Self::from),
+        }
+    }
+
+    fn lte_const(self, other: Number) -> TCResult<Self::Compare> {
+        match self {
+            Self::Dense(dense) => dense.lte_const(other).map(Self::from),
+            Self::Sparse(sparse) => sparse.lte_const(other).map(Self::from),
+        }
+    }
+
+    fn ne_const(self, other: Number) -> TCResult<Self::Compare> {
+        match self {
+            Self::Dense(dense) => dense.ne_const(other).map(Self::from),
+            Self::Sparse(sparse) => sparse.ne_const(other).map(Self::from),
+        }
+    }
+}
+
 #[async_trait]
 impl<FD, FS, D, T> TensorIO<D> for Tensor<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
 {
     type Txn = T;
@@ -565,10 +725,11 @@ where
 #[async_trait]
 impl<FD, FS, D, T> TensorDualIO<D, Self> for Tensor<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
 {
     type Txn = T;
@@ -585,10 +746,11 @@ where
 
 impl<FD, FS, D, T> TensorMath<D, Self> for Tensor<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
 {
     type Combine = Self;
@@ -615,6 +777,13 @@ where
         }
     }
 
+    fn pow(self, other: Self) -> TCResult<Self::Combine> {
+        match self {
+            Self::Dense(this) => this.pow(other),
+            Self::Sparse(this) => this.pow(other),
+        }
+    }
+
     fn sub(self, other: Self) -> TCResult<Self::Combine> {
         match self {
             Self::Dense(this) => this.sub(other),
@@ -623,12 +792,60 @@ where
     }
 }
 
-impl<FD, FS, D, T> TensorReduce<D> for Tensor<FD, FS, D, T>
+impl<FD, FS, D, T> TensorMathConst for Tensor<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
+    D::FileClass: From<TensorType>,
+{
+    type Combine = Self;
+
+    fn add_const(self, other: Number) -> TCResult<Self::Combine> {
+        match self {
+            Self::Dense(dense) => dense.add_const(other).map(Self::from),
+            Self::Sparse(sparse) => sparse.add_const(other).map(Self::from),
+        }
+    }
+
+    fn div_const(self, other: Number) -> TCResult<Self::Combine> {
+        match self {
+            Self::Dense(dense) => dense.div_const(other).map(Self::from),
+            Self::Sparse(sparse) => sparse.div_const(other).map(Self::from),
+        }
+    }
+
+    fn mul_const(self, other: Number) -> TCResult<Self::Combine> {
+        match self {
+            Self::Dense(dense) => dense.mul_const(other).map(Self::from),
+            Self::Sparse(sparse) => sparse.mul_const(other).map(Self::from),
+        }
+    }
+
+    fn pow_const(self, other: Number) -> TCResult<Self::Combine> {
+        match self {
+            Self::Dense(dense) => dense.pow_const(other).map(Self::from),
+            Self::Sparse(sparse) => sparse.pow_const(other).map(Self::from),
+        }
+    }
+
+    fn sub_const(self, other: Number) -> TCResult<Self::Combine> {
+        match self {
+            Self::Dense(dense) => dense.sub_const(other).map(Self::from),
+            Self::Sparse(sparse) => sparse.sub_const(other).map(Self::from),
+        }
+    }
+}
+
+impl<FD, FS, D, T> TensorReduce<D> for Tensor<FD, FS, D, T>
+where
+    D: Dir,
+    T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
 {
     type Txn = T;
@@ -665,10 +882,11 @@ where
 
 impl<FD, FS, D, T> TensorTransform for Tensor<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
 {
     type Broadcast = Self;
@@ -732,10 +950,11 @@ where
 #[async_trait]
 impl<FD, FS, D, T> TensorUnary<D> for Tensor<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
 {
     type Txn = T;
@@ -748,6 +967,13 @@ where
         }
     }
 
+    fn exp(&self) -> TCResult<Self::Unary> {
+        match self {
+            Self::Dense(dense) => dense.exp().map(Self::from),
+            Self::Sparse(sparse) => sparse.exp().map(Self::from),
+        }
+    }
+
     async fn all(self, txn: T) -> TCResult<bool> {
         match self {
             Self::Dense(dense) => dense.all(txn).await,
@@ -772,10 +998,11 @@ where
 
 impl<FD, FS, D, T, B> From<DenseTensor<FD, FS, D, T, B>> for Tensor<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     B: DenseAccess<FD, FS, D, T>,
     D::FileClass: From<TensorType>,
 {
@@ -786,10 +1013,11 @@ where
 
 impl<FD, FS, D, T, A> From<SparseTensor<FD, FS, D, T, A>> for Tensor<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     A: SparseAccess<FD, FS, D, T>,
     D::FileClass: From<TensorType>,
 {
@@ -801,10 +1029,12 @@ where
 #[async_trait]
 impl<FD, FS, D, T> de::FromStream for Tensor<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
+    T: Transaction<D>,
     D::FileClass: From<BTreeType> + From<TensorType>,
 {
     type Context = T;
@@ -835,10 +1065,12 @@ impl<FD, FS, D, T> TensorVisitor<FD, FS, D, T> {
 #[async_trait]
 impl<FD, FS, D, T> de::Visitor for TensorVisitor<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
+    T: Transaction<D>,
     D::FileClass: From<BTreeType> + From<TensorType>,
 {
     type Value = Tensor<FD, FS, D, T>;
@@ -874,10 +1106,11 @@ where
 #[async_trait]
 impl<'en, FD, FS, D, T> IntoView<'en, D> for Tensor<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
 {
     type Txn = T;
