@@ -1,4 +1,3 @@
-use std::convert::TryFrom;
 use std::fmt;
 
 use afarray::{Array, CoordBlocks, CoordMerge, Coords};
@@ -7,6 +6,7 @@ use futures::future::{self, TryFutureExt};
 use futures::stream::{self, StreamExt, TryStreamExt};
 use futures::try_join;
 use log::debug;
+use safecast::AsType;
 
 use tc_btree::Node;
 use tc_error::*;
@@ -78,6 +78,7 @@ pub enum SparseAccessor<FD, FS, D, T> {
     Broadcast(Box<SparseBroadcast<FD, FS, D, T, Self>>),
     Cast(Box<SparseCast<FD, FS, D, T, Self>>),
     Combine(Box<SparseCombinator<FD, FS, D, T, Self, Self>>),
+    CombineConst(Box<SparseConstCombinator<FD, FS, D, T, Self>>),
     CombineLeft(Box<SparseLeftCombinator<FD, FS, D, T, Self, Self>>),
     Dense(Box<DenseToSparse<FD, FS, D, T, DenseAccessor<FD, FS, D, T>>>),
     Expand(Box<SparseExpand<FD, FS, D, T, Self>>),
@@ -90,10 +91,11 @@ pub enum SparseAccessor<FD, FS, D, T> {
 
 impl<FD, FS, D, T> TensorAccess for SparseAccessor<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
 {
     fn dtype(&self) -> NumberType {
@@ -101,6 +103,7 @@ where
             Self::Broadcast(broadcast) => broadcast.dtype(),
             Self::Cast(cast) => cast.dtype(),
             Self::Combine(combine) => combine.dtype(),
+            Self::CombineConst(combine) => combine.dtype(),
             Self::CombineLeft(combine) => combine.dtype(),
             Self::Dense(dense) => dense.dtype(),
             Self::Expand(expand) => expand.dtype(),
@@ -117,6 +120,7 @@ where
             Self::Broadcast(broadcast) => broadcast.ndim(),
             Self::Cast(cast) => cast.ndim(),
             Self::Combine(combine) => combine.ndim(),
+            Self::CombineConst(combine) => combine.ndim(),
             Self::CombineLeft(combine) => combine.ndim(),
             Self::Dense(dense) => dense.ndim(),
             Self::Expand(expand) => expand.ndim(),
@@ -133,6 +137,7 @@ where
             Self::Broadcast(broadcast) => broadcast.shape(),
             Self::Cast(cast) => cast.shape(),
             Self::Combine(combine) => combine.shape(),
+            Self::CombineConst(combine) => combine.shape(),
             Self::CombineLeft(combine) => combine.shape(),
             Self::Dense(dense) => dense.shape(),
             Self::Expand(expand) => expand.shape(),
@@ -149,6 +154,7 @@ where
             Self::Broadcast(broadcast) => broadcast.size(),
             Self::Cast(cast) => cast.size(),
             Self::Combine(combine) => combine.size(),
+            Self::CombineConst(combine) => combine.size(),
             Self::CombineLeft(combine) => combine.size(),
             Self::Dense(dense) => dense.size(),
             Self::Expand(expand) => expand.size(),
@@ -164,10 +170,11 @@ where
 #[async_trait]
 impl<FD, FS, D, T> SparseAccess<FD, FS, D, T> for SparseAccessor<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
 {
     type Slice = Self;
@@ -182,6 +189,7 @@ where
             Self::Broadcast(broadcast) => broadcast.filled(txn).await,
             Self::Cast(cast) => cast.filled(txn).await,
             Self::Combine(combine) => combine.filled(txn).await,
+            Self::CombineConst(combine) => combine.filled(txn).await,
             Self::CombineLeft(combine) => combine.filled(txn).await,
             Self::Dense(dense) => dense.filled(txn).await,
             Self::Expand(expand) => expand.filled(txn).await,
@@ -198,6 +206,7 @@ where
             Self::Broadcast(broadcast) => broadcast.filled_at(txn, axes).await,
             Self::Cast(cast) => cast.filled_at(txn, axes).await,
             Self::Combine(combine) => combine.filled_at(txn, axes).await,
+            Self::CombineConst(combine) => combine.filled_at(txn, axes).await,
             Self::CombineLeft(combine) => combine.filled_at(txn, axes).await,
             Self::Dense(dense) => dense.filled_at(txn, axes).await,
             Self::Expand(expand) => expand.filled_at(txn, axes).await,
@@ -214,6 +223,7 @@ where
             Self::Broadcast(broadcast) => broadcast.filled_count(txn).await,
             Self::Cast(cast) => cast.filled_count(txn).await,
             Self::Combine(combine) => combine.filled_count(txn).await,
+            Self::CombineConst(combine) => combine.filled_count(txn).await,
             Self::CombineLeft(combine) => combine.filled_count(txn).await,
             Self::Dense(dense) => dense.filled_count(txn).await,
             Self::Expand(expand) => expand.filled_count(txn).await,
@@ -230,6 +240,7 @@ where
             Self::Broadcast(broadcast) => broadcast.slice(bounds).map(SparseAccess::accessor),
             Self::Cast(cast) => cast.slice(bounds).map(SparseAccess::accessor),
             Self::Combine(combinator) => combinator.slice(bounds).map(SparseAccess::accessor),
+            Self::CombineConst(combinator) => combinator.slice(bounds).map(SparseAccess::accessor),
             Self::CombineLeft(combinator) => combinator.slice(bounds).map(SparseAccess::accessor),
             Self::Dense(dense) => dense.slice(bounds).map(SparseAccess::accessor),
             Self::Expand(expand) => expand.slice(bounds).map(SparseAccess::accessor),
@@ -246,21 +257,35 @@ where
             Self::Broadcast(broadcast) => {
                 broadcast.transpose(permutation).map(SparseAccess::accessor)
             }
+
             Self::Cast(cast) => cast.transpose(permutation).map(SparseAccess::accessor),
+
             Self::Combine(combinator) => combinator
                 .transpose(permutation)
                 .map(SparseAccess::accessor),
+
+            Self::CombineConst(combinator) => combinator
+                .transpose(permutation)
+                .map(SparseAccess::accessor),
+
             Self::CombineLeft(combinator) => combinator
                 .transpose(permutation)
                 .map(SparseAccess::accessor),
+
             Self::Dense(dense) => dense.transpose(permutation).map(SparseAccess::accessor),
+
             Self::Expand(expand) => expand.transpose(permutation).map(SparseAccess::accessor),
+
             Self::Reduce(reduce) => reduce.transpose(permutation).map(SparseAccess::accessor),
+
             Self::Table(table) => table.transpose(permutation).map(SparseAccess::accessor),
+
             Self::Slice(slice) => slice.transpose(permutation).map(SparseAccess::accessor),
+
             Self::Transpose(transpose) => {
                 transpose.transpose(permutation).map(SparseAccess::accessor)
             }
+
             Self::Unary(unary) => unary.transpose(permutation).map(SparseAccess::accessor),
         }
     }
@@ -269,10 +294,11 @@ where
 #[async_trait]
 impl<FD, FS, D, T> SparseWrite<FD, FS, D, T> for SparseAccessor<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
 {
     async fn write_value(&self, txn_id: TxnId, coord: Coord, value: Number) -> TCResult<()> {
@@ -285,10 +311,11 @@ where
 
 impl<FD, FS, D, T> ReadValueAt<D> for SparseAccessor<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
 {
     type Txn = T;
@@ -298,6 +325,7 @@ where
             Self::Broadcast(broadcast) => broadcast.read_value_at(txn, coord),
             Self::Cast(cast) => cast.read_value_at(txn, coord),
             Self::Combine(combine) => combine.read_value_at(txn, coord),
+            Self::CombineConst(combine) => combine.read_value_at(txn, coord),
             Self::CombineLeft(combine) => combine.read_value_at(txn, coord),
             Self::Dense(dense) => dense.read_value_at(txn, coord),
             Self::Expand(expand) => expand.read_value_at(txn, coord),
@@ -316,6 +344,7 @@ impl<FD, FS, D, T> fmt::Display for SparseAccessor<FD, FS, D, T> {
             Self::Broadcast(broadcast) => fmt::Display::fmt(broadcast, f),
             Self::Cast(cast) => fmt::Display::fmt(cast, f),
             Self::Combine(combinator) => fmt::Display::fmt(combinator, f),
+            Self::CombineConst(combinator) => fmt::Display::fmt(combinator, f),
             Self::CombineLeft(combinator) => fmt::Display::fmt(combinator, f),
             Self::Dense(dense) => fmt::Display::fmt(dense, f),
             Self::Expand(expand) => fmt::Display::fmt(expand, f),
@@ -362,10 +391,11 @@ where
 #[async_trait]
 impl<FD, FS, D, T, B> SparseAccess<FD, FS, D, T> for DenseToSparse<FD, FS, D, T, B>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     B: DenseAccess<FD, FS, D, T>,
     D::FileClass: From<TensorType>,
 {
@@ -495,12 +525,13 @@ pub struct SparseBroadcast<FD, FS, D, T, A> {
 
 impl<FD, FS, D, T, A> SparseBroadcast<FD, FS, D, T, A>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node>,
     D: Dir,
     T: Transaction<D>,
-    A: SparseAccess<FD, FS, D, T>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
+    A: SparseAccess<FD, FS, D, T>,
 {
     pub fn new(source: A, shape: Shape) -> TCResult<Self> {
         debug!("SparseBroadcast::new {} into {}", source.shape(), shape);
@@ -541,12 +572,13 @@ where
 #[async_trait]
 impl<FD, FS, D, T, A> SparseAccess<FD, FS, D, T> for SparseBroadcast<FD, FS, D, T, A>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
-    A: SparseAccess<FD, FS, D, T>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
+    A: SparseAccess<FD, FS, D, T>,
 {
     type Slice = SparseBroadcast<FD, FS, D, T, A::Slice>;
     type Transpose = SparseTranspose<FD, FS, D, T, Self>;
@@ -576,10 +608,7 @@ where
         let coords = filled_at
             .map_ok(move |coords| stream::iter(coords.to_vec()).map(TCResult::Ok))
             .try_flatten()
-            .map_ok(move |coord| {
-                debug!("SparseBroadcast::filled at source coord {:?}", coord);
-                rebase.map_coord(coord)
-            })
+            .map_ok(move |coord| rebase.map_coord(coord))
             .map_ok(move |bounds| stream::iter(bounds.affected().map(TCResult::Ok)))
             .try_flatten();
 
@@ -664,12 +693,13 @@ where
 
 impl<FD, FS, D, T, A> ReadValueAt<D> for SparseBroadcast<FD, FS, D, T, A>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node>,
     D: Dir,
     T: Transaction<D>,
-    A: SparseAccess<FD, FS, D, T>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
+    A: SparseAccess<FD, FS, D, T>,
 {
     type Txn = T;
 
@@ -1027,6 +1057,138 @@ impl<FD, FS, D, T, L, R> fmt::Display for SparseCombinator<FD, FS, D, T, L, R> {
     }
 }
 
+#[derive(Clone)]
+pub struct SparseConstCombinator<FD, FS, D, T, A> {
+    source: A,
+    other: Number,
+    combinator: fn(Number, Number) -> Number,
+    phantom: Phantom<FD, FS, D, T>,
+}
+
+impl<FD, FS, D, T, A> SparseConstCombinator<FD, FS, D, T, A> {
+    pub fn new(source: A, other: Number, combinator: fn(Number, Number) -> Number) -> Self {
+        Self {
+            source,
+            other,
+            combinator,
+            phantom: Phantom::default(),
+        }
+    }
+}
+
+impl<FD, FS, D, T, A> TensorAccess for SparseConstCombinator<FD, FS, D, T, A>
+where
+    FD: File<Array>,
+    FS: File<Node>,
+    D: Dir,
+    T: Transaction<D>,
+    A: SparseAccess<FD, FS, D, T>,
+{
+    fn dtype(&self) -> NumberType {
+        Ord::max(self.source.dtype(), self.other.class())
+    }
+
+    fn ndim(&self) -> usize {
+        self.source.ndim()
+    }
+
+    fn shape(&'_ self) -> &'_ Shape {
+        self.source.shape()
+    }
+
+    fn size(&self) -> u64 {
+        self.source.size()
+    }
+}
+
+#[async_trait]
+impl<FD, FS, D, T, A> SparseAccess<FD, FS, D, T> for SparseConstCombinator<FD, FS, D, T, A>
+where
+    FD: File<Array>,
+    FS: File<Node>,
+    D: Dir,
+    T: Transaction<D>,
+    A: SparseAccess<FD, FS, D, T>,
+{
+    type Slice = SparseConstCombinator<FD, FS, D, T, A::Slice>;
+    type Transpose = SparseConstCombinator<FD, FS, D, T, A::Transpose>;
+
+    fn accessor(self) -> SparseAccessor<FD, FS, D, T> {
+        let this = SparseConstCombinator {
+            source: self.source.accessor(),
+            other: self.other,
+            combinator: self.combinator,
+            phantom: self.phantom,
+        };
+
+        SparseAccessor::CombineConst(Box::new(this))
+    }
+
+    async fn filled<'a>(self, txn: T) -> TCResult<SparseStream<'a>> {
+        let combinator = self.combinator;
+        let other = self.other;
+        let filled = self.source.filled(txn).await?;
+        Ok(Box::pin(filled.map_ok(move |(coord, value)| {
+            (coord, combinator(value, other))
+        })))
+    }
+
+    async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCBoxTryStream<'a, Coords>> {
+        self.source.filled_at(txn, axes).await
+    }
+
+    async fn filled_count(self, txn: T) -> TCResult<u64> {
+        self.source.filled_count(txn).await
+    }
+
+    fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
+        let slice = self.source.slice(bounds)?;
+        Ok(SparseConstCombinator::new(
+            slice,
+            self.other,
+            self.combinator,
+        ))
+    }
+
+    fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
+        let transpose = self.source.transpose(permutation)?;
+        Ok(SparseConstCombinator::new(
+            transpose,
+            self.other,
+            self.combinator,
+        ))
+    }
+}
+
+impl<FD, FS, D, T, A> ReadValueAt<D> for SparseConstCombinator<FD, FS, D, T, A>
+where
+    D: Dir,
+    T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    A: SparseAccess<FD, FS, D, T>,
+{
+    type Txn = T;
+
+    fn read_value_at<'a>(self, txn: T, coord: Coord) -> Read<'a> {
+        Box::pin(async move {
+            let combinator = self.combinator;
+            let other = self.other;
+
+            self.source
+                .read_value_at(txn, coord)
+                .map_ok(|(coord, val)| (coord, combinator(val, other)))
+                .await
+        })
+    }
+}
+
+impl<FD, FS, D, T, A> fmt::Display for SparseConstCombinator<FD, FS, D, T, A> {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        f.write_str("sparse constant combinator")
+    }
+}
+
 #[derive(Clone)]
 pub struct SparseLeftCombinator<FD, FS, D, T, L, R> {
     left: L,
@@ -1273,12 +1435,13 @@ where
 #[async_trait]
 impl<FD, FS, D, T, A> SparseAccess<FD, FS, D, T> for SparseExpand<FD, FS, D, T, A>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
-    A: SparseAccess<FD, FS, D, T>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
+    A: SparseAccess<FD, FS, D, T>,
 {
     type Slice = SparseAccessor<FD, FS, D, T>;
     type Transpose = SparseExpand<FD, FS, D, T, A::Transpose>;
@@ -1442,10 +1605,11 @@ pub struct SparseReduce<FD, FS, D, T> {
 
 impl<FD, FS, D, T> SparseReduce<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
 {
     pub fn new(
@@ -1463,10 +1627,11 @@ where
 
 impl<FD, FS, D, T> TensorAccess for SparseReduce<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
 {
     fn dtype(&self) -> NumberType {
@@ -1489,10 +1654,11 @@ where
 #[async_trait]
 impl<FD, FS, D, T> SparseAccess<FD, FS, D, T> for SparseReduce<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
 {
     type Slice = SparseReduce<FD, FS, D, T>;
@@ -1581,10 +1747,11 @@ where
 
 impl<FD, FS, D, T> ReadValueAt<D> for SparseReduce<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
 {
     type Txn = T;
@@ -1658,13 +1825,14 @@ where
 #[async_trait]
 impl<FD, FS, D, T, A> SparseAccess<FD, FS, D, T> for SparseTranspose<FD, FS, D, T, A>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
+    D::FileClass: From<TensorType>,
     A: SparseAccess<FD, FS, D, T>,
     A::Slice: SparseAccess<FD, FS, D, T>,
-    D::FileClass: From<TensorType>,
 {
     type Slice = <A::Slice as SparseAccess<FD, FS, D, T>>::Transpose;
     type Transpose = A::Transpose;
@@ -1790,10 +1958,11 @@ impl<FD, FS, D, T> SparseUnary<FD, FS, D, T> {
 
 impl<FD, FS, D, T> TensorAccess for SparseUnary<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
 {
     fn dtype(&self) -> NumberType {
@@ -1816,10 +1985,11 @@ where
 #[async_trait]
 impl<FD, FS, D, T> SparseAccess<FD, FS, D, T> for SparseUnary<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
 {
     type Slice = Self;
@@ -1865,10 +2035,11 @@ where
 
 impl<FD, FS, D, T> ReadValueAt<D> for SparseUnary<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
 {
     type Txn = T;
