@@ -1,4 +1,3 @@
-use std::convert::TryFrom;
 use std::fmt;
 use std::marker::PhantomData;
 use std::ops::{Add, Div, Mul, Sub};
@@ -10,26 +9,28 @@ use destream::{de, en};
 use futures::future::{self, TryFutureExt};
 use futures::stream::{self, Stream, StreamExt, TryStreamExt};
 use log::debug;
+use safecast::{AsType, CastFrom};
 
 use tc_btree::{BTreeType, Node};
 use tc_error::*;
 use tc_transact::fs::{CopyFrom, Dir, File, Hash, Persist, Restore};
 use tc_transact::{IntoView, Transact, Transaction, TxnId};
-use tc_value::{Number, NumberClass, NumberInstance, NumberType};
+use tc_value::{FloatType, Number, NumberClass, NumberInstance, NumberType};
 use tcgeneric::{TCBoxTryFuture, TCBoxTryStream};
 
 use super::dense::{BlockListSparse, DenseTensor, PER_BLOCK};
 use super::transform;
 use super::{
-    Bounds, Coord, Phantom, Schema, Shape, Tensor, TensorAccess, TensorBoolean, TensorCompare,
-    TensorDualIO, TensorIO, TensorInstance, TensorMath, TensorReduce, TensorTransform, TensorType,
-    TensorUnary,
+    Bounds, Coord, Phantom, Schema, Shape, Tensor, TensorAccess, TensorBoolean, TensorBooleanConst,
+    TensorCompare, TensorCompareConst, TensorDualIO, TensorIO, TensorInstance, TensorMath,
+    TensorMathConst, TensorReduce, TensorTransform, TensorType, TensorUnary, ERR_COMPLEX_EXPONENT,
 };
 
 use access::*;
 pub use access::{DenseToSparse, SparseAccess, SparseAccessor, SparseWrite};
 pub use table::SparseTable;
 
+#[allow(unused)]
 mod access;
 mod combine;
 mod table;
@@ -135,10 +136,11 @@ where
 
 impl<FD, FS, D, T> SparseTensor<FD, FS, D, T, SparseTable<FD, FS, D, T>>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<BTreeType>,
 {
     /// Create a new `SparseTensor` with the given schema
@@ -187,7 +189,6 @@ impl<FD, FS, D, T, A> TensorInstance for SparseTensor<FD, FS, D, T, A> {
     }
 }
 
-#[async_trait]
 impl<FD, FS, D, T, L, R> TensorBoolean<SparseTensor<FD, FS, D, T, R>>
     for SparseTensor<FD, FS, D, T, L>
 where
@@ -214,15 +215,15 @@ where
     }
 }
 
-#[async_trait]
 impl<FD, FS, D, T, A> TensorBoolean<Tensor<FD, FS, D, T>> for SparseTensor<FD, FS, D, T, A>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
-    A: SparseAccess<FD, FS, D, T>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
+    A: SparseAccess<FD, FS, D, T>,
 {
     type Combine = Tensor<FD, FS, D, T>;
     type LeftCombine = Tensor<FD, FS, D, T>;
@@ -249,16 +250,43 @@ where
     }
 }
 
+impl<FD, FS, D, T, A> TensorBooleanConst for SparseTensor<FD, FS, D, T, A>
+where
+    FD: File<Array>,
+    FS: File<Node>,
+    D: Dir,
+    T: Transaction<D>,
+    A: SparseAccess<FD, FS, D, T>,
+{
+    type Combine = SparseTensor<FD, FS, D, T, SparseConstCombinator<FD, FS, D, T, A>>;
+
+    fn and_const(self, other: Number) -> TCResult<Self::Combine> {
+        let access = SparseConstCombinator::new(self.accessor, other, Number::and);
+        Ok(access.into())
+    }
+
+    fn or_const(self, other: Number) -> TCResult<Self::Combine> {
+        let access = SparseConstCombinator::new(self.accessor, other, Number::or);
+        Ok(access.into())
+    }
+
+    fn xor_const(self, other: Number) -> TCResult<Self::Combine> {
+        let access = SparseConstCombinator::new(self.accessor, other, Number::xor);
+        Ok(access.into())
+    }
+}
+
 impl<FD, FS, D, T, L, R> TensorCompare<SparseTensor<FD, FS, D, T, R>>
     for SparseTensor<FD, FS, D, T, L>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
+    D::FileClass: From<TensorType>,
     L: SparseAccess<FD, FS, D, T>,
     R: SparseAccess<FD, FS, D, T>,
-    D::FileClass: From<TensorType>,
 {
     type Compare = SparseTensor<FD, FS, D, T, SparseCombinator<FD, FS, D, T, L, R>>;
     type Dense = Condensed<FD, FS, D, T, L, R>;
@@ -314,12 +342,13 @@ where
 
 impl<FD, FS, D, T, A> TensorCompare<Tensor<FD, FS, D, T>> for SparseTensor<FD, FS, D, T, A>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
-    A: SparseAccess<FD, FS, D, T>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
+    A: SparseAccess<FD, FS, D, T>,
 {
     type Compare = Tensor<FD, FS, D, T>;
     type Dense = Tensor<FD, FS, D, T>;
@@ -367,17 +396,70 @@ where
     }
 }
 
+impl<FD, FS, D, T, A> TensorCompareConst for SparseTensor<FD, FS, D, T, A> {
+    type Compare = SparseTensor<FD, FS, D, T, SparseConstCombinator<FD, FS, D, T, A>>;
+
+    fn eq_const(self, other: Number) -> TCResult<Self::Compare> {
+        fn eq(l: Number, r: Number) -> Number {
+            (l == r).into()
+        }
+
+        Ok(SparseConstCombinator::new(self.accessor, other, eq).into())
+    }
+
+    fn gt_const(self, other: Number) -> TCResult<Self::Compare> {
+        fn gt(l: Number, r: Number) -> Number {
+            (l > r).into()
+        }
+
+        Ok(SparseConstCombinator::new(self.accessor, other, gt).into())
+    }
+
+    fn gte_const(self, other: Number) -> TCResult<Self::Compare> {
+        fn gte(l: Number, r: Number) -> Number {
+            (l >= r).into()
+        }
+
+        Ok(SparseConstCombinator::new(self.accessor, other, gte).into())
+    }
+
+    fn lt_const(self, other: Number) -> TCResult<Self::Compare> {
+        fn lt(l: Number, r: Number) -> Number {
+            (l < r).into()
+        }
+
+        Ok(SparseConstCombinator::new(self.accessor, other, lt).into())
+    }
+
+    fn lte_const(self, other: Number) -> TCResult<Self::Compare> {
+        fn lte(l: Number, r: Number) -> Number {
+            (l <= r).into()
+        }
+
+        Ok(SparseConstCombinator::new(self.accessor, other, lte).into())
+    }
+
+    fn ne_const(self, other: Number) -> TCResult<Self::Compare> {
+        fn ne(l: Number, r: Number) -> Number {
+            (l != r).into()
+        }
+
+        Ok(SparseConstCombinator::new(self.accessor, other, ne).into())
+    }
+}
+
 #[async_trait]
 impl<FD, FS, D, T, L, R> TensorDualIO<D, SparseTensor<FD, FS, D, T, R>>
     for SparseTensor<FD, FS, D, T, L>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
+    D::FileClass: From<TensorType>,
     L: SparseWrite<FD, FS, D, T>,
     R: SparseAccess<FD, FS, D, T>,
-    D::FileClass: From<TensorType>,
 {
     type Txn = T;
 
@@ -415,12 +497,13 @@ where
 #[async_trait]
 impl<FD, FS, D, T, A> TensorDualIO<D, Tensor<FD, FS, D, T>> for SparseTensor<FD, FS, D, T, A>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
-    A: SparseWrite<FD, FS, D, T>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
+    A: SparseWrite<FD, FS, D, T>,
 {
     type Txn = T;
 
@@ -509,6 +592,15 @@ where
         self.left_combine(other, Number::mul)
     }
 
+    fn pow(self, other: SparseTensor<FD, FS, D, T, R>) -> TCResult<Self::LeftCombine> {
+        if !other.dtype().is_real() {
+            return Err(TCError::unsupported(ERR_COMPLEX_EXPONENT));
+        }
+
+        debug!("SparseTensor::pow");
+        self.left_combine(other, Number::pow)
+    }
+
     fn sub(self, other: SparseTensor<FD, FS, D, T, R>) -> TCResult<Self::Combine> {
         debug!("SparseTensor::sub");
         self.combine(other, Number::sub)
@@ -517,12 +609,13 @@ where
 
 impl<FD, FS, D, T, A> TensorMath<D, Tensor<FD, FS, D, T>> for SparseTensor<FD, FS, D, T, A>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
-    A: SparseAccess<FD, FS, D, T>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
+    A: SparseAccess<FD, FS, D, T>,
 {
     type Combine = Tensor<FD, FS, D, T>;
     type LeftCombine = Tensor<FD, FS, D, T>;
@@ -548,6 +641,13 @@ where
         }
     }
 
+    fn pow(self, other: Tensor<FD, FS, D, T>) -> TCResult<Self::Combine> {
+        match other {
+            Tensor::Sparse(sparse) => self.mul(sparse).map(Tensor::from),
+            Tensor::Dense(dense) => self.mul(dense.into_sparse()).map(Tensor::from),
+        }
+    }
+
     fn sub(self, other: Tensor<FD, FS, D, T>) -> TCResult<Self::Combine> {
         match other {
             Tensor::Sparse(sparse) => self.sub(sparse).map(Tensor::from),
@@ -556,14 +656,43 @@ where
     }
 }
 
+impl<FD, FS, D, T, A> TensorMathConst for SparseTensor<FD, FS, D, T, A> {
+    type Combine = SparseTensor<FD, FS, D, T, SparseConstCombinator<FD, FS, D, T, A>>;
+
+    fn add_const(self, other: Number) -> TCResult<Self::Combine> {
+        Ok(SparseConstCombinator::new(self.accessor, other, Number::add).into())
+    }
+
+    fn div_const(self, other: Number) -> TCResult<Self::Combine> {
+        Ok(SparseConstCombinator::new(self.accessor, other, Number::div).into())
+    }
+
+    fn mul_const(self, other: Number) -> TCResult<Self::Combine> {
+        Ok(SparseConstCombinator::new(self.accessor, other, Number::mul).into())
+    }
+
+    fn pow_const(self, other: Number) -> TCResult<Self::Combine> {
+        if !other.class().is_real() {
+            return Err(TCError::unsupported(ERR_COMPLEX_EXPONENT));
+        }
+
+        Ok(SparseConstCombinator::new(self.accessor, other, Number::pow).into())
+    }
+
+    fn sub_const(self, other: Number) -> TCResult<Self::Combine> {
+        Ok(SparseConstCombinator::new(self.accessor, other, Number::sub).into())
+    }
+}
+
 impl<FD, FS, D, T, A> TensorReduce<D> for SparseTensor<FD, FS, D, T, A>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
-    A: SparseAccess<FD, FS, D, T>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
+    A: SparseAccess<FD, FS, D, T>,
     Self: TensorInstance,
     <Self as TensorInstance>::Dense: TensorReduce<D, Txn = T> + Send + Sync,
 {
@@ -619,12 +748,13 @@ where
 
 impl<FD, FS, D, T, A> TensorTransform for SparseTensor<FD, FS, D, T, A>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node>,
     D: Dir,
     T: Transaction<D>,
-    A: SparseAccess<FD, FS, D, T>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
+    A: SparseAccess<FD, FS, D, T>,
 {
     type Broadcast = SparseTensor<FD, FS, D, T, SparseBroadcast<FD, FS, D, T, A>>;
     type Cast = SparseTensor<FD, FS, D, T, SparseCast<FD, FS, D, T, A>>;
@@ -678,6 +808,18 @@ where
         Ok(SparseTensor::from(accessor))
     }
 
+    fn exp(&self) -> TCResult<Self::Unary> {
+        fn exp(n: Number) -> Number {
+            let n = f64::cast_from(n);
+            n.exp().into()
+        }
+
+        let dtype = NumberType::Float(FloatType::F64);
+        let source = self.accessor.clone().accessor();
+        let accessor = SparseUnary::new(source, exp, dtype);
+        Ok(SparseTensor::from(accessor))
+    }
+
     async fn all(self, txn: Self::Txn) -> TCResult<bool> {
         let affected = stream::iter(Bounds::all(self.shape()).affected());
         let filled = self.accessor.filled(txn).await?;
@@ -710,12 +852,13 @@ where
 impl<FD, FS, D, T, A> CopyFrom<D, SparseTensor<FD, FS, D, T, A>>
     for SparseTensor<FD, FS, D, T, SparseTable<FD, FS, D, T>>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
-    A: SparseAccess<FD, FS, D, T>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<BTreeType> + From<TensorType>,
+    A: SparseAccess<FD, FS, D, T>,
 {
     async fn copy_from(
         instance: SparseTensor<FD, FS, D, T, A>,
@@ -748,10 +891,11 @@ where
 #[async_trait]
 impl<FD, FS, D, T> Persist<D> for SparseTensor<FD, FS, D, T, SparseTable<FD, FS, D, T>>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<BTreeType> + From<TensorType>,
 {
     type Schema = Schema;
@@ -772,10 +916,11 @@ where
 #[async_trait]
 impl<FD, FS, D, T> Restore<D> for SparseTensor<FD, FS, D, T, SparseTable<FD, FS, D, T>>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<BTreeType> + From<TensorType>,
 {
     async fn restore(&self, backup: &Self, txn_id: TxnId) -> TCResult<()> {
@@ -816,12 +961,13 @@ impl<FD, FS, D, T, A> fmt::Display for SparseTensor<FD, FS, D, T, A> {
 #[async_trait]
 impl<'en, FD, FS, D, T, A> IntoView<'en, D> for SparseTensor<FD, FS, D, T, A>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node>,
     D: Dir,
     T: Transaction<D>,
-    A: SparseAccess<FD, FS, D, T>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
+    A: SparseAccess<FD, FS, D, T>,
 {
     type Txn = T;
     type View = SparseTensorView<'en>;
@@ -840,10 +986,11 @@ where
 #[async_trait]
 impl<FD, FS, D, T> de::FromStream for SparseTensor<FD, FS, D, T, SparseTable<FD, FS, D, T>>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<BTreeType> + From<TensorType>,
 {
     type Context = T;
@@ -874,10 +1021,11 @@ impl<FD, FS, D, T> SparseTensorVisitor<FD, FS, D, T> {
 #[async_trait]
 impl<FD, FS, D, T> de::Visitor for SparseTensorVisitor<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<BTreeType> + From<TensorType>,
 {
     type Value = SparseTensor<FD, FS, D, T, SparseTable<FD, FS, D, T>>;
